// This example is not self-contained.
// It requires usage of the example driver specific to your platform.
// See the HAL documentation.

#include "src/bindings/hal_common.h"
#include "src/bindings/bricklet_rs485.h"

void check(int rc, const char *msg);
void example_setup(TF_HAL *hal);
void example_loop(TF_HAL *hal);

static uint8_t received_request_id = 0;
static bool register_address_valid = false;

// Callback function for Modbus slave write single register request callback
static void modbus_slave_write_single_register_request_handler(TF_RS485 *device, uint8_t request_id,
                                                               uint32_t register_address,
                                                               uint16_t register_value,
                                                               void *user_data) {
	(void)device; (void)user_data; // avoid unused parameter warning

	tf_hal_printf("Request ID: %I8u\n", request_id);
	tf_hal_printf("Register Address: %I32u\n", register_address);
	tf_hal_printf("Register Value: %I16u\n", register_value);
	tf_hal_printf("\n");

	received_request_id = request_id;
	register_address_valid = register_address == 42;

	if (!register_address_valid) {
		tf_hal_printf("Error: Invalid register address\n");
	}
}

static TF_RS485 rs485;

void example_setup(TF_HAL *hal) {
	// Create device object
	check(tf_rs485_create(&rs485, NULL, hal), "create device object");

	// Set operating mode to Modbus RTU slave
	check(tf_rs485_set_mode(&rs485, TF_RS485_MODE_MODBUS_SLAVE_RTU), "call set_mode");

	// Modbus specific configuration:
	// - slave address = 17
	// - master request timeout = 0ms (unused in slave mode)
	check(tf_rs485_set_modbus_configuration(&rs485, 17,
	                                        0), "call set_modbus_configuration");

	// Register Modbus slave write single register request callback to function
	// modbus_slave_write_single_register_request_handler
	tf_rs485_register_modbus_slave_write_single_register_request_callback(&rs485,
	                                                                      modbus_slave_write_single_register_request_handler,
	                                                                      NULL);
}

void example_loop(TF_HAL *hal) {
	// Poll for callbacks
	// Polling with 0 will process one packet at most, so we can't miss a request.
	tf_hal_callback_tick(hal, 0);

	if(received_request_id == 0) {
		// We didn't receive a request.
		return;
	}

	if(!register_address_valid) {
		tf_rs485_modbus_slave_report_exception(&rs485, received_request_id, TF_RS485_EXCEPTION_CODE_ILLEGAL_DATA_ADDRESS);
	} else {
		tf_rs485_modbus_slave_answer_write_single_register_request(&rs485, received_request_id);
	}

	// Make sure we respond to the request only once.
	received_request_id = 0;
}
